from typing import Literal, Optional, Sequence

import numpy as np
import pandas as pd

from .species import Species

from gprofiler import GProfiler
import aiohttp
import asyncio
import nest_asyncio

# for JupyterNotebook support
nest_asyncio.apply()


class DIOPTRelease:
    """
    Util class for fetching orthologs.
    """

    gp = GProfiler(return_dataframe=True)

    def __init__(
        self, version: Literal["v8", "v9"] = "v8", input_species: str | int = "human"
    ) -> None:
        if version not in ["v8", "v9"]:
            if version in [8, 9]:
                self.version = "v" + version
            else:
                raise TypeError("The")
        else:
            if version == "v9":
                print(
                    "Please note that v9 is currently in beta testing. Use v8 if you prefer the stable version."
                )
            self.version = version

        self.input_species = Species.parse_species(input_species)

    def fetch(
        self,
        genes: Sequence[str],
        target_species: Optional[Species | str | int],
        filter: Optional[
            Literal["best_match", "exclude_score_less_1", "exclude_score_less_2"]
        ] = None,
        condensed: bool = True,
    ) -> pd.DataFrame:
        """
        Queries the DIOPT RESTful API endpoint for orthologs.

        You can theoretically fetch information for all annotated species by passing in None for target species, but just so you know it will be a lot slower.

        By default, only select information will be returned in the output pd.DataFrame. If comprehensive information is desired, pass in condensed=False. Note this action is both memory and time consuming.
        """

        def url_builder(
            api_version: str,
            input_species: int,
            target_species: Optional[int],
            filter: str,
        ):
            return (
                lambda entrez_gene_id: f"https://www.flyrnai.org/tools/diopt/web/diopt_api/{api_version}/get_orthologs_from_entrez/{input_species}/{entrez_gene_id}/{target_species if target_species else 'all'}/{filter}"
            )

        target_species_id = (
            Species.parse_species(target_species).ncbi_taxonomy_id
            if target_species
            else "all"
        )

        base = url_builder(
            self.version,
            self.input_species.ncbi_taxonomy_id,
            target_species_id,
            filter if filter else "none",
        )

        genes = self.convert_to_entrez_acc(np.array(genes).tolist())

        async def _fetch(session, url):
            async with session.get(url) as response:
                try:
                    return await response.json()
                except:
                    raise ValueError(
                        f"The returned data cannot be deserialized.\nQuery: {url}\nResponse: {response}"
                    )

        async def _fetch_all(urls):
            async with aiohttp.ClientSession() as session:
                tasks = []
                for url in urls:
                    task = asyncio.create_task(_fetch(session, url))
                    tasks.append(task)
                results = await asyncio.gather(*tasks)
                return results

        results = asyncio.run(_fetch_all([base(gene) for gene in genes]))

        unmatched = []
        chunks = []

        for entry in results:
            gene_details = entry["search_details"]["gene_details"][0]

            if not entry["results"]:
                print(entry)
                unmatched.append(gene_details["symbol"])
                continue

            for i in entry["results"].values():
                if condensed:
                    chunk = pd.DataFrame(i.values())[
                        [
                            "species_id",
                            "symbol",
                            "geneid",
                            "species_specific_geneid",
                            "species_specific_geneid_type",
                            "confidence",
                        ]
                    ]
                    chunk = chunk.add_prefix("target_")
                    chunk.insert(0, "input_symbol", gene_details["symbol"])
                    chunk.insert(
                        1, "input_ensembl_geneid", gene_details["ensembl_geneid"]
                    )
                else:
                    chunk = pd.DataFrame(i.values())
                    chunk = chunk.add_prefix("target_")
                    chunk = chunk.assign(
                        **{f"input_{key}": value for key, value in gene_details.items()}
                    )
                chunks.append(chunk)

        if unmatched:
            print(f"Some of the queries are returned unmatched:\n{unmatched}")

        return pd.concat(chunks, axis=0, ignore_index=True)

    def convert_to_entrez_acc(self, genes: Sequence[str]) -> np.array:
        """
        Helper function that converts a list of arbitrary gene identifiers to Entrez IDs using the gprofiler library.

        Theoretically the list of genes can be in any format but it should preferabbly be passed in as a list, np.ndarray, or pd.Series.

        Returns an np.array of the Entrez IDs. If the full pd.DataFrame is desired, use the gprofiler library directly.
        """
        organism = self.input_species.latin_name.split(" ")
        organism = organism[0][0] + organism[1]

        result = self.gp.convert(
            organism=organism.lower(),
            query=np.array(genes).tolist(),
            target_namespace="ENTREZGENE_ACC",
        )

        if result[result["converted"] == "None"].shape[0]:
            print(
                f'The following queries returned None when trying to get Entrez Acc Number:\n{result[result["converted"] == "None"]["incoming"].tolist()}'
            )

        return result[result["converted"] != "None"]["converted"].to_numpy()



from typing import Optional, Union


class Species:
    def __init__(
        self,
        latin_name: str,
        common_name: Optional[str | list[str]],
        ncbi_taxonomy_id: int,
    ) -> None:
        self.latin_name = latin_name
        self.common_name = common_name
        self.ncbi_taxonomy_id = ncbi_taxonomy_id
        self.registered: bool = False

    def __str__(self) -> str:
        return self.latin_name

    latin_name_map: dict[str, "Species"] = {}
    common_name_map: dict[str, "Species"] = {}
    ncbi_taxonomy_id_map: dict[int, "Species"] = {}

    @staticmethod
    def Register(species: "Species") -> None:
        Species.latin_name_map[species.latin_name] = species
        if species.common_name:
            if type(species.common_name) == list:
                for common_name in species.common_name:
                    Species.common_name_map[common_name] = species
            else:
                Species.common_name_map[species.common_name] = species
        Species.ncbi_taxonomy_id_map[species.ncbi_taxonomy_id] = species

        species.registered = True

    @staticmethod
    def parse_species(
        search_query: Union["Species", str, int], error: bool = False
    ) -> "Species":
        """
        Parameters:
            search_query: str | int | Species - any input that points to a species
            error: bool - whether to throw error if unable to parse species from search_query
        """

        if type(search_query) == int:
            if search_query in Species.ncbi_taxonomy_id_map:
                return Species.ncbi_taxonomy_id_map[search_query]
            else:
                message = f"NCBI Taxonomy id not found: {search_query}. A list of available species can be retrived by calling Species.available()."
                if error:
                    raise TypeError(message)
                else:
                    print(message)
        else:
            # if it is a registered species, return the species directly
            if type(search_query) == Species and search_query.registered:
                return search_query

            search_string = str(search_query).lower()

            # try matching common names first
            if search_string in Species.common_name_map:
                return Species.common_name_map[search_string]

            # try matching latin names directly
            search_string = search_string.capitalize()

            if search_string in Species.latin_name_map:
                return Species.common_name_map[search_string]

            # try matching abbreviated latin names
            species_query = search_string.split(" ")[-1]
            for genus_species in Species.latin_name_map.keys():
                if species_query == genus_species.split(" ")[-1]:
                    return Species.latin_name_map[genus_species]

            message = f"Name not found: {search_query}. A list of available species can be retrived by calling Species.available()"
            if error:
                raise TypeError(message)
            else:
                print(message)

    @staticmethod
    def available() -> None:
        print("Available species through DIOPT:")
        for ncbi_taxnomoy_id, species in Species.ncbi_taxonomy_id_map.items():
            print(f"{ncbi_taxnomoy_id}\t{species.latin_name}")


Species.Register(Species("Escherichia coli", common_name=None, ncbi_taxonomy_id=83333))
Species.Register(Species("Arabidopsis thaliana", ["thale cress", "cress"], 3702))
Species.Register(Species("Schizosaccharomyces pombe", "fission yeast", 4896))
Species.Register(Species("Saccharomyces cerevisiae", ["budding yeast", "yeast"], 4932))
Species.Register(Species("Caenorhabditis elegans", ["worm", "nematode"], 6239))
Species.Register(Species("Anopheles gambiae", "mosquito", 7165))
Species.Register(Species("Drosophila melanogaster", ["fly", "fruit fly"], 7227))
Species.Register(Species("Danio rerio", "zebrafish", 7955))
Species.Register(Species("Xenopus tropicalis", "western clawed frog", 8364))
Species.Register(Species("Rattus norvegicus", ["rat", "brown rat"], 10116))
Species.Register(Species("Mus musculus", "mouse", 10090))
Species.Register(Species("Homo sapiens", "human", 9606))



"""
__version__.py
~~~~~~~~~~~~~~

Information about the current version of the pyDIOPT package.
"""

__title__ = 'pyDIOPT'
__description__ = 'pyDIOPT - Python Wrapper for DIOPT'
__version__ = '0.1.0'
__author__ = 'AZ Zhang'
__author_email__ = 'azz@berkeley.edu'
__license__ = 'Apache 2.0'
__url__ = 'https://github.com/akmazian/pyDIOPT'



def citation() -> None:
    print("Please cite the original DIOPT papers")
    print(
        """
        Yanhui Hu, Ian Flockhart, Arunachalam Vinayagam, Clemens Bergwitz, Bonnie Berger, Norbert Perrimon, and Stephanie E Mohr. 2011. “An integrative approach to ortholog prediction for disease-focused and other functional studies.” BMC Bioinformatics, 12, Pp. 357.
        
        Yanhui Hu, Aram Comjean, Jonathan Rodiger, Yifang Liu, Yue Gao, Verena Chung, Jonathan Zirin, Norbert Perrimon, and Stephanie E Mohr. 2020. “FlyRNAi.org-the database of the Drosophila RNAi screening center and transgenic RNAi project: 2021 update.” Nucleic Acids Res.
        """
    )
    print()

    print("The links can be found here:")
    print(
        "[BMC Bioinformatics](https://fgr.hms.harvard.edu/publications/integrative-approach-ortholog-prediction-disease-focused-and-other-functional)"
    )
    print(
        "[Nucleic Acids Res](https://fgr.hms.harvard.edu/publications/flyrnaiorg-database-drosophila-rnai-screening-center-and-transgenic-rnai-project)"
    )

    print()

    print(
        "It would be appreciated if you also cited pyDIOPT :) though it is not required."
    )


def help() -> None:
    print("The documentations for pyDIOPT can be found at:")
    print("*not quite ready yet*")
    print()
    print("The documentations for DIOPT can be found at:")
    print("[DIOPT](https://fgr.hms.harvard.edu/diopt-documentation)")



#!/usr/bin/env python3

import os
from setuptools import setup

# get key package details from py_pkg/__version__.py
about = {}  # type: ignore
here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, "pyDIOPT", "__version__.py")) as f:
    exec(f.read(), about)

# load the README file and use it as the long_description for PyPI
with open("README.md", "r") as f:
    readme = f.read()

# package configuration - for reference see:
# https://setuptools.readthedocs.io/en/latest/setuptools.html#id9
setup(
    name=about["__title__"],
    description=about["__description__"],
    long_description=readme,
    long_description_content_type="text/markdown",
    version=about["__version__"],
    author=about["__author__"],
    author_email=about["__author_email__"],
    url=about["__url__"],
    packages=["pyDIOPT"],
    include_package_data=True,
    python_requires=">=3.10",
    install_requires=[
        "numpy",
        "requests",
        "pandas",
        "twine",
        "aiohttp",
        "gprofiler-official",
    ],
    license=about["__license__"],
    zip_safe=False,
    classifiers=[
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "Programming Language :: Python :: 3.11",
    ],
    keywords="wrapper bioinformatics biology",
)
